#include "ReedSolomon.h"
#include <iostream>


#include "schifra_galois_field.hpp"
#include "schifra_galois_field_polynomial.hpp"
#include "schifra_sequential_root_generator_polynomial_creator.hpp"
#include "schifra_reed_solomon_encoder.hpp"
#include "schifra_reed_solomon_decoder.hpp"
#include "schifra_reed_solomon_block.hpp"
#include "schifra_error_processes.hpp"
#include "schifra_reed_solomon_bitio.hpp"



char* Encode(std::string sendMessage)

{

    /* Finite Field Parameters */

    const std::size_t field_descriptor                = 4;

    const std::size_t generator_polynomial_index      = 0;

    const std::size_t generator_polynomial_root_count = 7;

    /* Reed Solomon Code Parameters */

    const std::size_t code_length = 15; //(2^4 - 1)

    const std::size_t fec_length  =  7;

    const std::size_t data_length = code_length - fec_length;

    const schifra::galois::field field(field_descriptor,

                                       schifra::galois::primitive_polynomial_size01,

                                       schifra::galois::primitive_polynomial01);

    schifra::galois::field_polynomial generator_polynomial(field);

    schifra::make_sequential_root_generator_polynomial(field, generator_polynomial_index,

                                                              generator_polynomial_root_count,

                                                              generator_polynomial);

    typedef schifra::reed_solomon::encoder<code_length,fec_length> encoder_t;

    const encoder_t encoder(field,generator_polynomial);

    char* tmp = new char[fec_length];
    for(int i=0;i<fec_length;i++)
      tmp[i]=0x0;
    schifra::reed_solomon::block<code_length,fec_length> block(sendMessage, tmp);

    if(!encoder.encode(block)) std::cout <<block.error<<std::endl ;

    std::string dataString(data_length,' ');

    block.data_to_string(dataString);

    std::string fecString(fec_length,' ');

    block.fec_to_string(fecString);

    // Initialize the result

    char* result = new char[code_length];
    for(int i=0;i<code_length;i++)
	result[i]=0x0;

    // Add the data string to the result

    for(size_t i=0;i<data_length;i++)

        result[i]=dataString[i];

    // Add the FEC string to the result

    for(size_t i=0;i<fec_length;i++){

        result[i+data_length]=fecString[i];
    }

    // Return the result
    return result;

}

std::string Decode(char* buffer)

{

    /* Finite Field Parameters */

    const std::size_t field_descriptor                = 4;

    const std::size_t generator_polynomial_index      = 0;

    const std::size_t generator_polynomial_root_count = 7;

    /* Reed Solomon Code Parameters */

    const std::size_t code_length = 15; //(2^4 - 1)

    const std::size_t fec_length  =  7;

    const std::size_t data_length = code_length - fec_length;

    const schifra::galois::field field(field_descriptor,

                                       schifra::galois::primitive_polynomial_size01,

                                       schifra::galois::primitive_polynomial01);

    schifra::galois::field_polynomial generator_polynomial(field);

    schifra::make_sequential_root_generator_polynomial(field, generator_polynomial_index,

                                                              generator_polynomial_root_count,

                                                              generator_polynomial);

    /* Instantiate Encoder and Decoder (Codec) */

    typedef schifra::reed_solomon::decoder<code_length,fec_length> decoder_t;

    const decoder_t decoder(field,generator_polynomial_index);

    // Split the buffer in two strings

    std::string dataString(data_length,(char)0x0);

    std::string fecString(fec_length,(char)0x0);

    for(size_t i=0;i<data_length;i++)

        dataString[i]= buffer[i];

    for(size_t i=0;i<fec_length;i++)

        fecString[i]= buffer[i+data_length];

    // Generate the block

    schifra::reed_solomon::block<code_length,fec_length> block(dataString,fecString);

    // Decode the block

    decoder.decode(block);

    // Return the decoded message

    std::string resultString(data_length,(char)0x0);

    block.data_to_string(resultString);

    return resultString;

}
