#include "Interleaver.h"
#include "ReedSolomon.h"
#include "viterbi.h"
#include <string>
#include <iostream>
#include <iomanip>
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdint.h>
#include <fstream>

std::string ViterbiEncode(std::string data);
std::string ViterbiDecoder(std::string data);
std::string BitsToStingOfBits(char* buffer, int length);
char* StringToBits(std::string bitString, int length);

#define heigth  15
#define width   2

int fdr, fdw;
struct {
    char c1;
    char c2;
    char c3;
    char c4;	
}tologic, fromlogic;


int main(int argc, char** argv)
{
    fdw = open("/dev/xillybus_write_32", O_WRONLY);
	if (fdw <0) 
	{
	    perror("Failed to open Xillybus device file(s)");
	    exit(1);
	}
	if (argc != 2)
	{
		std::cout << "Not enough input arguments" << std::endl;
		return -1;
	}
   std::string message = argv[1];

   if(message.length()<4 || message.length()>8)
	{
std::cout << "Not the correct length" << std::endl;
return -1;
	}
	

  //std::string message = "xillinux";
  std::cout << "Original: ";
  std::cout << message << std::endl << std::endl;

  //Reed Solomon encode
  char* RS = Encode(message);
 /* std::cout << "Reed Solomon: ";
  for (int i = 0; i < 15; i++) std::cout << "0x" << std::hex << (((int)RS[i]) & 0x00FF) << " ";
  std::cout << std::endl << std::endl;*/

  //Interleaver
  char* Inter = new char[heigth*width]();
  FillInterleaver(RS, 15, Inter, heigth, width);
  char* InterRes = Interleave(Inter, heigth, width);
  // Print the values in the array
  /*std::cout << "Interleaver: ";
 for (int i = 0; i < 30; i++) std::cout << "0x" << std::hex << (((int)InterRes[i]) & 0x00FF) << " ";    // Print two new lines
  std::cout << std::endl << std::endl;*/

  //Viterbi Encode
  std::string ViterbiIn = BitsToStingOfBits(InterRes, heigth*width);
  std::string ViterbiOut = ViterbiEncode(ViterbiIn);
 // std::cout << "Viterbi: ";
 // std::cout << ViterbiOut << std::endl << std::endl;

   // 93: fixed length result of viterbi is 738 characters (0 / 1 character)
   	char* transResult = StringToBits(ViterbiOut, 93);
   	for(int i=0;i<93;i+=4)
	{
		tologic.c1=transResult[i];
		if(i<92)
		{
			tologic.c2=transResult[i+1];
			tologic.c3=transResult[i+2];
			tologic.c4=transResult[i+3];
		}
		else
		{
			tologic.c2=0x0;
			tologic.c3=0x0;
			tologic.c4=0x0;
		}

		write(fdw, (void *) &tologic, sizeof(tologic));
	}


}
std::string BitsToStingOfBits(char* buffer, int length)
{
  // Generate a result string, 8 times the number of bits
  std::string Result(length * 8, ' ');
  // Count the number of characters
  int counter = 0;

  for (int byteCounter = 0; byteCounter < length; byteCounter++)
  {
    for (int bitCounter = 0; bitCounter < 8; bitCounter++)
    {
      // Generate a mask, with only a 1 at the position of the expected bit
      unsigned char mask = 0x80 >> bitCounter;

      // Use the mask to get the bit out of the current byte
      unsigned char bitValue = buffer[byteCounter] & mask;

      // If the bit value is 0 return 0
      if (bitValue == 0x00)
        Result[counter] = '0';
      else if (bitValue >> (7-bitCounter) == 0x01)
      {
        Result[counter] = '1';
      }
      else
      {
        Result[counter] = 'X';
      }
      // Increment the counter
      counter++;
    }
  }
  return Result;
}

char* StringToBits(std::string bitString, int length)
{
  // Count the number of characters
  int counter = 0;

  // Initialize the result buffer
  char* buffer = new char[length]();

  for(int i=0;i<length;i++)
	buffer[i]=0;

  // For each byte in the result
  for (int byteCounter = 0; byteCounter < length; byteCounter++)
  {
    // For each bit the result
    for (int bitCounter = 0; bitCounter < 8; bitCounter++)
    {
      // Declare the bit value as zero
      unsigned char bitValue = 0x0;

      // If the character is a one
      if (bitString[counter] == '1')
      {
        // Set the bit value to 1
        bitValue = 0x01;
      }
      // Move the bit to its correct position
      bitValue = bitValue << (7 - bitCounter);

      // Add the bit to the buffer
      buffer[byteCounter] = buffer[byteCounter] | bitValue;
      counter ++;
    }
  }
  return buffer;
}

std::string ViterbiEncode(std::string data)
{
	const int constraint = 7;
	std::vector<int> polynomials;
	polynomials.push_back(91);
	polynomials.push_back(121);
	polynomials.push_back(117);
	const std::string& bits = data;
	ViterbiCodec codec(constraint, polynomials);
	std::string encodeddata = codec.Encode(bits);
	return encodeddata;
}
std::string ViterbiDecoder(std::string data)
{
	const int constraint = 7;
	std::vector<int> polynomials;
	polynomials.push_back(91);
	polynomials.push_back(121);
	polynomials.push_back(117);
	const std::string& bits = data;
	ViterbiCodec codec(constraint, polynomials);
	std::string encodeddata = codec.Decode(bits);
	return encodeddata;
}
